#ifndef SRC_RODSOUP_CC_INCLUDED
#define SRC_RODSOUP_CC_INCLUDED
/* Autogenerated from src/rodsoup.cc - changes will be overwritten. */

#include "common.hh"

class Vector
{
public:
    float x, y;

    Vector () {}
    Vector (float x_, float y_) : x(x_), y(y_)
    {
    }

    float dot (Vector v) {return x * v.x + y * v.y;}
    float cross (Vector v) {return x * v.y - y * v.x;}
    float len () {return sqrtf(x * x + y * y);}
    Vector add (Vector v) {return Vector(x + v.x, y + v.y);}
    Vector sub (Vector v) {return Vector(x - v.x, y - v.y);}
    Vector scale (float s) {return Vector(x * s, y * s);}
    Vector operator + (Vector v) {return add(v);}
    Vector operator - (Vector v) {return sub(v);}
    float operator ^ (Vector v) {return cross(v);}
    float operator | (Vector v) {return dot(v);}
    Vector operator * (float s) {return scale(s);}
};

class Rod
{
public:
    float x, y;
    float dx, dy;
    RGB rgb;
    Rod *next, *prev;

    void init()
    {
        next = NULL;
        prev = NULL;
    }

    Rod (float x_, float y_, float dx_, float dy_, RGB rgb_) :
        x(x_), y(y_), dx(dx_), dy(dy_), rgb(rgb_)
    {
        init();
    }

    bool is_left(Vector &v)
    {
        Vector a(x - v.x, y - v.y);
        return (a ^ v) > 0.0;
    }
    bool is_right(Vector &v)
    {
        Vector a(x - v.x, y - v.y);
        return (a ^ v) < 0.0;
    }

    bool Rod::intersection (Rod const *rod, Vector *inter);
    bool Rod::intersection2(Rod const *rod, Vector *inter);
};

class Soup
{
public:
    Rod *first, *last;
    Soup()
    {
        first = NULL;
        last = NULL;
    }

    ~Soup()
    {
        Rod *next;
        for (Rod *r = first; r; r = next)
        {
            next = r->next;
            delete r;
        }
    }

    void add(Rod *rod)
    {
        rod->next = NULL;
        rod->prev = last;
        if (last)
            last->next = rod;
        else
            first = rod;

        last = rod;
    }

    void add_random (int n);
};

#ifndef _CPROTO_
#include "rodsoup.pro"
#endif /* _CPROTO_ */
#endif /* SRC_RODSOUP_CC_INCLUDED */
