#ifdef _PROTOTYPE_

#include "common.hh"
#define GLYPH_TARGET GLYPH_TARGET_ALLEGGL
#include <glyph.h>

class TheFont
{
public:
    TheFont(char const *filename_, int size_);
    GLYPH_FACE *face;
    GLYPH_REND *rend;
    GLYPH_TEXTURE *texture;

    char *filename;
    int size;

    void set();

    static void print(char const *text, ...);
    static void write(char const *text, ...);
    static void text_color(float r, float g, float b, float a);
    static void text_pos(float x, float y);

    static void reload();
    void reload_me();

    static TheFont *first;
    TheFont *next;
};

extern float text_x;
extern float text_y;

#endif /* _PROTOTYPE_ */

#include "font.hh"

TheFont *TheFont::first = NULL;

float text_x = 0;
float text_y = 0;
static TheFont *text_font = NULL;
float text_w = 0;
float text_h = 0;

static int workaround_bug_size = 16;

void TheFont::reload_me()
{
    face = gk_load_face_from_file(filename, 0);
    rend = gk_create_renderer(face, NULL);
    gk_rend_set_size_pixels(rend, size, size);
    gk_rend_set_text_color_rgb(rend, 0, 0, 0);
    gk_rend_set_text_alpha(rend, 255);
    text_font = this;

    texture = gk_create_texture(rend, 32, 96); 
}

TheFont::TheFont(char const *filename_, int size_)
{
    filename = strdup(filename_);
    size = size_;
    reload_me();
    next = first;
    first = this;
}

void TheFont::reload()
{
    TheFont *f;
    gk_library_cleanup();
    for (f = first; f; f = f->next)
    {
        f->reload_me();
    }
}

void TheFont::set()
{
    text_font = this;
}

void TheFont::text_color(float r, float g, float b, float a)
{
    TheFont *self = text_font;
    gk_rend_set_text_color_rgb(self->rend, char(r * 255), char(g * 255), char(b * 255));
    gk_rend_set_text_alpha(self->rend, char(a * 255));
}

void TheFont::text_pos(float x, float y)
{
    text_x = x;
    text_y = y;
}

static void _print(char const *text, int newline)
{
    TheFont *self = text_font;
    int w = 2 + gk_text_width_utf8(self->rend, text);
    int h = 2 + gk_text_height_utf8(self->rend, text);
    int a = (gk_face_ascender(self->face) >> 6) * h / workaround_bug_size;
    gk_render_line_gl_utf8(self->texture, text, int(text_x + 3), int(text_y + a));
    if (newline)
    {
        text_y += h;
    }
    else
    {
        text_x += w;
    }
    text_w = w;
    text_h = h;
}

void TheFont::print(char const *text, ...)
{
    char str[1024];
    va_list args;
    va_start(args, text);
    vsnprintf(str, sizeof str, text, args);
    va_end(args);
    _print(str, 1);
}

void TheFont::write(char const *text, ...)
{
    char str[1024];
    va_list args;
    va_start(args, text);
    vsnprintf(str, sizeof str, text, args);
    va_end(args);
    _print(str, 0);
}
