#ifdef _PROTOTYPE_
extern class Runner *my_runner;
extern int level;
extern class TheDisplay *display;
#endif /* _PROTOTYPE_ */

#include <common.hh>

#include "display.hh"
#include "runner.hh"
#include "rodsoup.hh"
#include "font.hh"
#include "image.hh"
#include "title.hh"

TheDisplay *display;

class Tile
{
public:
    float light;
    int tile;
};

class Photon : public Vector
{
public:
    Photon(float x_, float y_) : Vector(x_, y_)
    {
        alive = false;
        tile_x = 0;
        tile_y = 0;
        dir = Vector(0, 0);
        ticks = 0;
    }

    bool alive;
    Vector dir;
    int ticks;

    int tile_x, tile_y;
};

class Draggable
{
public:
    Draggable(int x_, int y_) : x(x_), y(y_)
    {
        rods[0] = NULL;
        rods[1] = NULL;
        rods[2] = NULL;
        rods[3] = NULL;
    }
    Rod *rods[4];
    int x, y;
};

class Runner *my_runner;
int level = 1;
TitleRunner *title_runner;

class MainRunner : public Runner
{
public:
    Soup *soup;
    Soup *out;

    Image *tile[256];
    Image *shadow;
    Image *sparkle;

    SAMPLE *ding, *tak, *tok, *woop, *woosh, *zoing, *woo;

    static int const LIGHTS = 20;
    float LIGHT_RADIUS;
    static int const FPS= 60;

    Photon *light[LIGHTS];

    int draggables_count;
    Draggable *draggable[100];

    Vector cross;

    int drag;

    int mx, my, mz, omx, omy, omz;
    int mb, omb;

    int light_time;
    int light_next_time;
    int lights_created;
    int starters;
    int start_x[10], start_y[10];
    int lights_in_target;
    int lights_lost;
    int lights_in_level;

    int out_timer;

    Tile *map;

#define FLAT 1
#define BOX_ 2
#define STAR 3
#define GOAL 4
#define SLAS 5
#define BSLA 6
#define MBOX 7
#define MBSL 8
#define MSLA 9
#define FBSL 10
#define FSLA 11

    TheFont *small;

    MainRunner(char const *name_) : Runner(name_)
    {
        tile[FLAT] = new Image("data/pics/flat.png");
        tile[BOX_] = new Image("data/pics/box.png");
        tile[STAR] = new Image("data/pics/start.png");
        tile[GOAL] = new Image("data/pics/goal.png");
        tile[SLAS] = new Image("data/pics/slash.png");
        tile[BSLA] = new Image("data/pics/backslash.png");
        tile[MBOX] = new Image("data/pics/mbox.png");
        tile[MBSL] = new Image("data/pics/mbackslash.png");
        tile[MSLA] = new Image("data/pics/mslash.png");
        tile[FBSL] = new Image("data/pics/fbackslash.png");
        tile[FSLA] = new Image("data/pics/fslash.png");
        shadow = new Image("data/pics/shadow.png");
        sparkle = new Image("data/pics/sparkle.png");

        map = new Tile[30 * 40];

        out = new Soup();
        int B = -10;
        RGB rgb = {0, 0, 0, 0};
        out->add(new Rod(B, SCREEN_H - B, 0, -SCREEN_H + B + B, rgb));
        out->add(new Rod(SCREEN_W - B, SCREEN_H - B, -SCREEN_W + B + B, 0, rgb));
        out->add(new Rod(SCREEN_W - B, B, 0, SCREEN_H - B - B, rgb));
        out->add(new Rod(B, B, SCREEN_W - B - B, 0, rgb));

        int i;
        for (i = 0; i < LIGHTS; i++)
        {
            light[i] = new Photon(0, 0);
        }

        small = new TheFont("data/fonts/font.ttf", 30);

        ding = load_sample("data/sounds/ding.wav");
        tak = load_sample("data/sounds/tak.wav");
        tok = load_sample("data/sounds/tok.wav");
        woop = load_sample("data/sounds/woop.wav");
        woosh = load_sample("data/sounds/woosh.wav");
        zoing = load_sample("data/sounds/zoing.wav");
        woo = load_sample("data/sounds/woo.wav");
    }

    void enter(void)
    {
        load_level();
    }

    void cleanup(void)
    {
        int i;
        for (i = 0; i < draggables_count; i++)
        {
            delete draggable[i];
        }
        if (soup)
            delete soup;
    }

    void load_level(void)
    {
        lights_in_target = 0;
        LIGHT_RADIUS = 150.0;
        int i;
        out_timer = 0;
        drag = -1;
        light_time = FPS * 1;
        light_next_time = FPS;
        lights_created = 0;
        soup = new Soup();
        lights_lost = 0;
        draggables_count = 0;
        int lookup[256];
        char lname[1024];
        sprintf(lname, "data/level%d.txt", level);
        PACKFILE *pack = pack_fopen(lname, "r");
        int size = file_size(lname);
        unsigned char *textbuffer = new unsigned char[size];
        unsigned char *buffer = textbuffer;
        pack_fread(buffer, size, pack);
        pack_fclose(pack);
        starters = 0;
        int x, y;
        int S = 16;
        lookup[(int)'x'] = BOX_;
        lookup[(int)'s'] = STAR;
        lookup[(int)'g'] = GOAL;
        lookup[(int)'/'] = SLAS;
        lookup[(int)'\\'] = BSLA;
        lookup[(int)','] = MSLA;
        lookup[(int)'`'] = MBSL;
        lookup[(int)'o'] = MBOX;
        lookup[(int)'%'] = FSLA;
        lookup[(int)'#'] = FBSL;
        for (y = 0; y < 30; y++)
        {
            for (x = 0; x < 40; x++)
            {
                map[y * 40 + x].tile = 1;
                //int r, g, b;
                //hsv_to_rgb(rand() % 360, 1, 0.5, &r, &g, &b);
                RGB rgb = {32, 32, 32, 128};
                if (*buffer == 's')
                {
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                    if (*buffer == 's')
                    {
                        start_x[starters] = x * S + 8;
                        start_y[starters] = SCREEN_H - y * S - 8;
                        starters++;
                    }
                }
                else if (*buffer == 'g')
                {
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                }
                else if (*buffer == 'x' || *buffer == 'o')
                {
                    draggable[draggables_count] = new Draggable(x, y);
                    Draggable *d = draggable[draggables_count];
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                    soup->add(d->rods[0] = new Rod(x * S, SCREEN_H - y * S,
                        S, 0, rgb));
                    soup->add(d->rods[1] = new Rod(x * S + S, SCREEN_H - y * S,
                        0, -S, rgb));
                    soup->add(d->rods[2] = new Rod(x * S + S, SCREEN_H - y * S - S,
                        -S, 0, rgb));
                    soup->add(d->rods[3] = new Rod(x * S, SCREEN_H - y * S - S,
                        0, S, rgb));
                    draggables_count++;
                    if (*buffer == 'x')
                    {
                        draggables_count--;
                        delete d;
                    }
                }
                else if (*buffer == '#' || *buffer == '%')
                {
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                }
                else if (*buffer == '\\' || *buffer == '`')
                {
                    draggable[draggables_count] = new Draggable(x, y);
                    Draggable *d = draggable[draggables_count];
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                    soup->add(d->rods[0] = new Rod(x * S, SCREEN_H - y * S,
                        S, -S, rgb));
                    soup->add(d->rods[1] = new Rod(x * S + S, SCREEN_H - y * S - S,
                        -S, S, rgb));
                    draggables_count++;
                    if (*buffer != '`')
                    {
                        draggables_count--;
                        delete d;
                    }
                }
                else if (*buffer == '/' || *buffer == ',')
                {
                    draggable[draggables_count] = new Draggable(x, y);
                    Draggable *d = draggable[draggables_count];
                    map[y * 40 + x].tile = lookup[(int)*buffer];
                    soup->add(d->rods[0] = new Rod(x * S, SCREEN_H - y * S - S,
                        S, S, rgb));
                    soup->add(d->rods[1] = new Rod(x * S + S, SCREEN_H - y * S,
                        -S, -S, rgb));
                    draggables_count++;
                    if (*buffer != ',')
                    {
                        draggables_count--;
                        delete d;
                    }
                }
                map[y * 40 + x].light = 1;
                buffer++;
            }
            while (*buffer < ' ')
                buffer++;
        }
        lights_in_level = 0;
        while (*buffer == '+')
        {
            lights_in_level++;
            buffer++;
        }
        while (*buffer == '-')
        {
            light_next_time += FPS;
            buffer++;
        }
        delete[] textbuffer;

        int s = 0;
        for (i = 0; i < LIGHTS; i++)
        {
            light[i]->dir = Vector(1, 0);
            light[i]->x = start_x[s];
            light[i]->y = start_y[s];
            light[i]->alive = false;
            s++;
            if (s == starters)
                s = 0;
        }
    }

    int find_nearest(float x, float y, float mind)
    {
        int i, j = -1;
        for (i = 0; i < draggables_count; i++)
        {
            Draggable *d = draggable[i];
            if (int(x / 16) == d->x && int(y / 16) == 29 - d->y)
            {
                return i;
            }
        }
        return j;
    }

    void drag_it(int i, float x, float y)
    {
        Draggable *d = draggable[i];
        int dx = int(x / 16);
        int dy = 29 - int(y / 16);
        if (dx != d->x || dy != d->y)
        {
            if (map[dy * 40 + dx].tile == 1)
            {
                int t = map[d->y * 40 + d->x].tile;
                map[d->y * 40 + d->x].tile = 1;
                dx -= d->x;
                dy -= d->y;
                d->x += dx;
                d->y += dy;
                int r;
                for (r = 0; r < 4; r++)
                {
                    if (d->rods[r])
                    {
                        d->rods[r]->x += dx * 16;
                        d->rods[r]->y -= dy * 16;
                    }
                }
                map[d->y * 40 + d->x].tile = t;
                play_sample(tak, 255, 128, 1000, 0);
            }
        }
    }

    void mouse_tick(void)
    {
        omx = mx;
        omy = my;
        omz = mz;
        omb = mb;
        mx = mouse_x;
        my = SCREEN_H - mouse_y;
        mz = mouse_z;
        mb = mouse_b;
    }

    int arrow_x(void)
    {
        return mx;
    }

    int arrow_y(void)
    {
        return my;
    }

    int arrow_z(void)
    {
        return mz;
    }

    int arrow_b(void)
    {
        return mb;
    }

    int arrow_delta_x(void)
    {
        return mx - omx;
    }

    int arrow_delta_y(void)
    {
        return my - omy;
    }

    int arrow_delta_z(void)
    {
        return mz - omz;
    }

    int arrow_delta_b(void)
    {
        return mb ^ omb;
    }

    void tick()
    {
        int i;
        mouse_tick();

        while (keypressed())
        {
            int k = readkey() >> 8;
            if (k == KEY_ESC)
            {
                if (key[KEY_C]) /* cheat */
                    level++;
                title_runner->switch_to();
                return;
            }
        }

        if (arrow_b() && arrow_delta_b())
        {
            drag = find_nearest(arrow_x(), arrow_y(), 10);
        }

        if (drag >= 0 && arrow_b())
        {
            drag_it(drag, arrow_x(), arrow_y());
        }

        if (lights_created < lights_in_level)
        {
            if (light_time)
                light_time--;
            else
            {
                light[lights_created]->alive = 1;
                lights_created++;
                light_time = light_next_time;
                play_sample(woosh, 255, 128, 1000, 0);
            }
        }

        if (out_timer)
        {
            out_timer--;
            LIGHT_RADIUS *= 0.99;
            if (!out_timer)
            {
                title_runner->switch_to();
                return;
            }
        }

        for (i = 0; i < LIGHTS; i++)
        {
            if (light[i]->alive)
            {
                light[i]->x += light[i]->dir.x;
                light[i]->y += light[i]->dir.y;

                int ix = int(light[i]->x / 16);
                int iy = int(light[i]->y / 16);
                int t = map[(29 - iy) * 40 + ix].tile;
                if (ix != light[i]->tile_x || iy != light[i]->tile_y)
                {
                    if (ix < 0 || iy < 0 || ix >= 40 || iy >= 30)
                    {
                        light[i]->alive = false;
                        lights_lost++;
                        if (!out_timer)
                        {
                            out_timer = FPS * 3;
                            play_sample(woop, 255, 128, 1000, 0);
                        }
                    }
                    else if (t == MBSL || t == BSLA || t == FBSL)
                    {
                        if ((Vector(light[i]->x - ix * 16, light[i]->y - iy * 16) - Vector(8, 8)).len()
                            <= 1)
                        {
                            light[i]->tile_x = ix;
                            light[i]->tile_y = iy;

                            float dx = -light[i]->dir.y;
                            float dy = -light[i]->dir.x;
                            light[i]->dir.x = dx;
                            light[i]->dir.y = dy;
                            if (t == FBSL)
                            {
                                map[(29 - iy) * 40 + ix].tile = FSLA;
                                play_sample(tok, 255, 128, 1000, 0);
                            }
                            play_sample(ding, 255, 128, 1500, 0);
                        }
                    }
                    else if (t == MSLA || t == SLAS || t == FSLA)
                    {
                        if ((Vector(light[i]->x - ix * 16, light[i]->y - iy * 16) - Vector(8, 8)).len()
                            <= 1)
                        {
                            light[i]->tile_x = ix;
                            light[i]->tile_y = iy;

                            float dx = light[i]->dir.y;
                            float dy = light[i]->dir.x;
                            light[i]->dir.x = dx;
                            light[i]->dir.y = dy;
                             if (t == FSLA)
                            {
                                map[(29 - iy) * 40 + ix].tile = FBSL;
                                play_sample(tok, 255, 128, 1000, 0);
                            }
                            play_sample(ding, 255, 128, 1500, 0);
                        }
                    }
                    else if (t == GOAL)
                    {
                        if ((Vector(light[i]->x - ix * 16, light[i]->y - iy * 16) - Vector(8, 8)).len()
                            <= 1)
                        {
                            light[i]->alive = false;
                            lights_in_target++;
                            play_sample(zoing, 255, 128, 1000, 0);
                            if (lights_in_target == lights_in_level)
                            {
                                level++;
                                out_timer = 1;
                                play_sample(woo, 255, 128, 1000, 0);
                            }
                        }
                    }
                    else if (t == MBOX || t == BOX_)
                    {
                        light[i]->dir.x = -light[i]->dir.x;
                        light[i]->dir.y = -light[i]->dir.y;
                        play_sample(ding, 255, 128, 1000, 0);
                    }
                    else
                    {
                        light[i]->tile_x = ix;
                        light[i]->tile_y = iy;
                    }
                }
            }
        }
    }

    Vector *current_light;
    void shadow_vertex(float x, float y)
    {
        glTexCoord2f((Vector(x, y) - *current_light).len() / 128.0, 0);
        glVertex2d(x, y);
    }

    void get_projection(Vector &point, Rod &rod)
    {
        current_light = &point;
        //Rod *r;
        //Vector a_, b_;
        Vector la(rod.x - point.x, rod.y - point.y);
        Vector lb(rod.x + rod.dx - point.x, rod.y + rod.dy - point.y);
        RGB rgb = {0, 0, 0, 0};
        //Rod ar(point.x, point.y, la.x, la.y, rgb);
        //Rod br(point.x, point.y, lb.x, lb.y, rgb);

        float side = la ^ Vector(rod.dx, rod.dy);
        const float epsilon1 = 0.001;
        if (side > -epsilon1)
            return;

        float l = LIGHT_RADIUS - (Vector(rod.x + rod.dx * 0.5, rod.y + rod.dy * 0.5) - point).len();
        if (l < 0)
            return;

        glColor4ub(rod.rgb.r, rod.rgb.g, rod.rgb.b, rod.rgb.filler);

        glBegin(GL_POLYGON);
        shadow_vertex(rod.x + la.x / la.len(), rod.y + la.y / la.len());
        shadow_vertex(rod.x + rod.dx + lb.x / lb.len(), rod.y + rod.dy + lb.y / lb.len());

        //bool ia = false, ib = false;

        shadow_vertex(rod.x + rod.dx + l * lb.x / lb.len(), rod.y + rod.dy + l * lb.y / lb.len());

        Rod ro(point.x, point.y, -rod.dy, rod.dx, rgb);
        Vector c;
        if (rod.intersection2(&ro, &c))
        {
            Vector cv(c.x - point.x, c.y - point.y);
            shadow_vertex(c.x + cv.x * l / cv.len(), c.y + cv.y * l / cv.len());
        }

        shadow_vertex(rod.x + l * la.x / la.len(), rod.y + l * la.y / la.len());

        /*
        if (side > 0)
            r = out->first;
        else
            r = out->last;
        for (; r;)
        {
            ib = r->intersection2(&br, &b_);
            if (ib)
            {
                shadow_vertex(b_.x, b_.y);
                break;
            }
            if (side > 0)
                r = r->next;
            else
                r = r->prev;
        }
        if (!ib)
        {
            b_.x = brln.x + brln.dx;
            b_.y = brln.y + brln.dy;
        }

        Rod *r2 = r;
        for (r = out->first; r; r = r->next)
        {
            if (!r2)
            {
                if (side > 0)
                    r2 = out->first;
                else
                    r2 = out->last;
            }
            ia = r2->intersection2(&ar, &a_);
            if (ia)
            {
                shadow_vertex(a_.x, a_.y);
                break;
            }
            if (side > 0)
            {
                shadow_vertex(r2->x, r2->y);
                r2 = r2->next;
            }
            else
            {

                shadow_vertex(r2->x + r2->dx, r2->y + r2->dy);
                r2 = r2->prev;
            }
        }*/

        glEnd();
    }

    float get_light(float x, float y)
    {
        int i;
        float l = 0;
        for (i = 0; i < LIGHTS; i++)
        {
            if (light[i]->alive)
            {
                float d = 1 - (Vector(x, y) - *light[i]).len() / LIGHT_RADIUS;
                if(d > 0)
                    l += d;
            }
        }
        glColor3f(l, l, l);
        return l;
    }

    void draw()
    {
        int i;

        Rod *rod;

        glLoadIdentity();

        float factor = 1;
        glScalef(factor, factor, 1);
        glTranslatef(0.5 * SCREEN_W * (1 - factor), 0.5 * SCREEN_H * (1 - factor), 0);

        glClearColor(1, 1, 1, 1);
        //glClearStencil(0);
        //glStencilMask(255);

        glClear(GL_COLOR_BUFFER_BIT);

        glEnable(GL_TEXTURE_2D);

        int xi, yi;
        for (yi = 0; yi < 30; yi++)
        {
            for (xi = 0; xi < 40; xi++)
            {
                int tl = map[yi * 40 + xi].tile;
                int x = xi * 16;
                int y = SCREEN_H - 16 - yi * 16;
                if (tl)
                {
                    glPushMatrix();
                    glTranslatef(x, y, 0);
                    glBindTexture(GL_TEXTURE_2D, tile[tl]->gl_texture);
                    glBegin(GL_TRIANGLE_FAN);

                    GLfloat w = tile[tl]->memory_cache->w;
                    GLfloat h = tile[tl]->memory_cache->h;

                    GLfloat l = 0;
                    GLfloat t = 0;
                    GLfloat r = tile[tl]->memory_cache->w;
                    GLfloat b = tile[tl]->memory_cache->h;

                    GLfloat mdx = l + tile[tl]->xoff;
                    GLfloat mdy = t + tile[tl]->yoff;

                    l += 0;
                    t += 0;
                    r -= 0;
                    b -= 0;

                    get_light(x + r, y + b);
                    glTexCoord2f(r / w, (h - b) / h);
                    glVertex2d(r - mdx, b - mdy);

                    get_light(x, y + b);
                    glTexCoord2f(l / w, (h - b) / h);
                    glVertex2d(l - mdx, b - mdy);

                    get_light(x, y);
                    glTexCoord2f(l / w, (h - t) / h);
                    glVertex2d(l - mdx, t - mdy);

                    get_light(x + r, y);
                    glTexCoord2f(r / w, (h - t) / h);
                    glVertex2d(r - mdx, t - mdy);

                    glEnd();
                    glPopMatrix();
                }
            }
        }

        /*
        glBegin(GL_LINES);
        glColor3f(1, 0, 0);
        for (rod = out->first; rod; rod = rod->next)
        {
            glVertex2d(rod->x, rod->y);
            glVertex2d(rod->x + rod->dx, rod->y + rod->dy);
        }
        glEnd();*/

        /*glBegin(GL_LINES);
        glColor4f(1, 1, 1, 1);
        for (rod = soup->first; rod; rod = rod->next)
        {
            glVertex2d(rod->x, rod->y);
            glVertex2d(rod->x + rod->dx, rod->y + rod->dy);
        }
        glEnd();*/

        glEnable(GL_TEXTURE_2D);
        //glDisable(GL_BLEND);
        //glDrawBuffer(GL_BACK);
        glBindTexture(GL_TEXTURE_2D, shadow->gl_texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

        /*glEnable(GL_STENCIL_TEST);
        glStencilMask(255);
        glStencilFunc(GL_GEQUAL, 0, 255);
        glStencilOp(GL_INCR, GL_INCR, GL_INCR);
*/
        for (i = 0; i < LIGHTS; i++)
        {
            if (light[i]->alive)
            {
                for (rod = soup->first; rod; rod = rod->next)
                {
                    get_projection(*light[i], *rod);
                }
            }
        }
/*
        glDrawBuffer(GL_BACK);
        glEnable(GL_BLEND);
        glStencilFunc(GL_GEQUAL, 0, 255);
        glStencilOp(GL_ZERO, GL_ZERO, GL_ZERO);
        glStencilMask(0);
*/
        glEnable(GL_TEXTURE_2D);
        for (i = 0; i < LIGHTS; i++)
        {
            if (light[i]->alive)
            {
                sparkle->xoff = 16;
                sparkle->yoff = 16;
                sparkle->draw_scaled_rotated_tinted(light[i]->x, light[i]->y,
                    fabs(sin(light[i]->ticks / 5.0)), fabs(cos(light[i]->ticks / 5.0)),
                    AL_PI * light[i]->ticks / (float)(FPS / 2), 1, 1, 1);
                light[i]->ticks += rand() & 1;
            }
        }

        glDisable(GL_TEXTURE_2D);

        glColor4f(1, 1, 0, 0.9);
        glPointSize(4);
        glBegin(GL_POINTS);
        glVertex2d(mx, my);
        glEnd();

        small->set();
        TheFont::text_color(1, 1, 1, 0.8);
        TheFont::text_pos(0, 0);
        TheFont::print("%d / %d", lights_in_target, lights_in_level);
    }
};

int main(void)
{
    allegro_init();

    install_mouse();
    install_keyboard();
    install_timer();
    loadpng_init();
    install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);

    srand(time(NULL));

    display = new TheDisplay(640, 480, 32, 120, TheDisplay::DISPLAY_FULLSCREEN);

    show_os_cursor(MOUSE_CURSOR_ARROW);

    my_runner = new MainRunner("main");
    title_runner = new TitleRunner("title");
    title_runner->switch_to();

    Runner::run();

    return 0;
}
END_OF_MAIN()
